import os
import sys
import asyncio
from dataclasses import dataclass
from distutils.util import strtobool
import logging

import discord
from discord.ext import commands, tasks
from dotenv import load_dotenv
import youtube_dl

from error import ErrorHandler
from message import NowPlayingMessage, QueuedMessage, ErrorMessage
from log import create_logger, DiscordLogger

load_dotenv()

# Suppress noise about console usage from errors
youtube_dl.utils.bug_reports_message = lambda: ''


@dataclass
class Song:
  title: str
  webpage_url: str
  audio_url: str


class Music(commands.Cog):
  def __init__(self, bot: commands.Bot, logger: logging.Logger = None):
    self._bot = bot
    self._queue = asyncio.Queue()
    self._queue_lock = asyncio.Lock()
    self._current_skip_message = None
    self._ytdl_params = {
        'cookiefile': os.environ.get('YOUTUBE_COOKIES'),
        'format': 'bestaudio/best',
        'outtmpl': '%(extractor)s-%(id)s-%(title)s.%(ext)s',
        'restrictfilenames': True,
        'noplaylist': True,
        'nocheckcertificate': True,
        'ignoreerrors': False,
        'logtostderr': False,
        'quiet': True,
        'no_warnings': True,
        'default_search': 'auto',
        'source_address': '0.0.0.0'  # bind to ipv4 since ipv6 addresses cause issues sometimes
    }
    self._ffmpeg_options = {
        'before_options': '-reconnect 1 -reconnect_streamed 1 -reconnect_delay_max 5',
        'options': '-vn'
    }
    self._ytdl = youtube_dl.YoutubeDL(self._ytdl_params)
    # pylint: disable=no-member
    self._handle_playback.start()

    self.logger = logger if logger else create_logger("bot")
    if log_channel_id := os.environ.get("DISCORD_LOG_CHANNEL_ID", None):
      clear_on_startup = os.environ.get("DISCORD_LOG_CHANNEL_CLEAR_ON_STARTUP", False)
      if clear_on_startup == "":
        clear_on_startup = False
      if isinstance(clear_on_startup, str):
        clear_on_startup = bool(strtobool(clear_on_startup))
      self.logger.addHandler(DiscordLogger(self._bot, channel_id=log_channel_id, clear_on_startup=clear_on_startup))

  def cog_unload(self):
    # pylint: disable=no-member
    self._handle_playback.cancel()

  def _next(self):
    """Trigger playback of next song."""
    self.logger.info("Song finished. Triggering playback of next song")
    self._queue.task_done()
    self._current_skip_message = None
    if self._queue_lock.locked():
      self._queue_lock.release()

  @tasks.loop()
  async def _handle_playback(self):
    while True:
      try:
        self.logger.info("Waiting for queue lock to acquire ...")
        await self._queue_lock.acquire()
        self.logger.info("Queue lock acquired! Waiting for song queue to return a song ...")
        ctx, song = await self._queue.get()
        self.logger.info("Queue returned a song!")
        if ctx.voice_client is None:
          # Bot is no longer in a voice channel.
          # This could be the case because a stop command was issued.
          # We will skip this (and possibly all remaining songs) in the queue
          self.logger.info('Bot is no longer in a voice channel. Skipping song "%s"', song.title)
          self._next()
          continue
        audio = discord.FFmpegPCMAudio(song.audio_url, **self._ffmpeg_options)

        def after(err):
          if err:
            self.logger.error("Player error: %s", err)
          self._next()
        self.logger.info('Now playing song "%s"', song.title)
        ctx.voice_client.play(audio, after=after)
        embed = NowPlayingMessage(title=song.title, url=song.webpage_url)
        msg = await ctx.send(embed=embed)
        await self._add_skip_button(msg)
      # pylint: disable=broad-except
      except Exception as err:
        self.logger.error("Error during playback: %s", err)
        if ctx:
          embed = ErrorMessage(str(err))
          await ctx.send(embed=embed)
        self._next()

  async def _add_skip_button(self, msg):
    await msg.add_reaction("⏭️")
    self._current_skip_message = msg

  @commands.Cog.listener()
  async def on_reaction_add(self, reaction, user):
    if not user.bot \
            and self._current_skip_message \
            and reaction.message.id == self._current_skip_message.id \
            and reaction.emoji == "⏭️":
      voice_client = reaction.message.guild.voice_client
      self._skip(voice_client)

  @_handle_playback.before_loop
  async def before_handle_playback(self):
    await self._bot.wait_until_ready()

  @commands.command()
  async def play(self, ctx, *, query):
    async with ctx.typing():
      data = self._ytdl.extract_info(query, download=False)
      if 'entries' in data:
        data = data['entries'][0]
        song = Song(
            title=data.get('title'),
            audio_url=data.get('url'),
            webpage_url=data.get('webpage_url')
        )
        self.logger.info('Qeueing song: title="%s", audio_url=%s, webpage_url=%s', song.title, song.audio_url, song.webpage_url)
        await self._queue.put((ctx, song))
        if ctx.voice_client.is_playing():
          embed = QueuedMessage(title=song.title, url=song.webpage_url)
          await ctx.send(embed=embed)

  def _skip(self, voice_client):
    """Skip to next song."""
    if voice_client is None or not voice_client.is_playing():
      raise commands.CommandError("No song playing")
    # This skips to next song because the bot does not differentiate between
    # a song stopping because it is finished or because it was manually stopped.
    self.logger.info("Skipping song")
    voice_client.stop()

  @commands.command()
  async def skip(self, ctx):
    async with ctx.typing():
      self._skip(ctx.voice_client)

  @commands.command()
  async def stop(self, ctx):
    self.logger.info("Stopping playback")
    await ctx.voice_client.disconnect()

  @play.before_invoke
  async def ensure_voice(self, ctx):
    if ctx.voice_client is None:
      if ctx.author.voice:
        self.logger.info("Connecting to voice channel ...")
        await ctx.author.voice.channel.connect()
        self.logger.info("Connected")
      else:
        raise commands.CommandError("Author not connected to a voice channel")


if __name__ == "__main__":
  prefix = os.environ.get("DISCORD_CMD_PREFIX", "!")
  logger = create_logger("bot")
  bot = commands.Bot(command_prefix=commands.when_mentioned_or(prefix),
                     description='Relatively simple music bot example')

  @bot.event
  async def on_ready():
    logger.info("Logged in as %s (%s)", bot.user, bot.user.id)
    logger.info('------')

  bot.add_cog(Music(bot, logger=logger))
  bot.add_cog(ErrorHandler(bot, logger=logger))

  token = os.environ.get("DISCORD_BOT_TOKEN", None)
  if not token:
    logger.warning("Discord bot token not found")
    sys.exit(1)

  bot.run(token)
